# include <stdio.h>
# include <math.h>
#include <algorithm>
#include <chrono>
#include <iostream>
#include<vector>
# include "autodiff/autodiff/forward/dual.hpp"
using namespace std;
using namespace std::chrono;
using namespace autodiff;

double f(double x)
{
    return x * sin( x * x);
}
double fp(double x)
{
    return sin(x*x)+ x* (2.0 * x) * cos(x * x);
}

dual af(dual x)
{
    return x * sin( x* x);
}

double fpfinite(double x)
{
    const double eps = 1e-8;
    return (f(x+eps)-f(x))/eps;
}

void printRealGradients(int n,double a,double b)
{
    auto start1 = high_resolution_clock::now();
    FILE *ft=fopen("realgradients.txt","w");
    for(double x= a;x<=b;x+=(b-a)/n)
    {
        fprintf(ft,"%20.10lf %20.10lf\n",x,fp(x));
    }
    fclose(ft);
    auto stop1 = high_resolution_clock::now();
    auto duration1 = duration_cast<microseconds>(stop1 - start1);
    printf("Time Passed: %ld \n",duration1);
}

void    printFiniteGradients(int n,double a,double b)
{
    auto start1 = high_resolution_clock::now();

    FILE *ft=fopen("finitegradients.txt","w");
    for(double x= a;x<=b;x+=(b-a)/n)
    {
        fprintf(ft,"%20.10lf %20.10lf\n",x,fpfinite(x));
    }
    fclose(ft);
    auto stop1 = high_resolution_clock::now();
    auto duration1 = duration_cast<microseconds>(stop1 - start1);
    printf("Time Passed: %ld \n",duration1);
}


void    printAutodiffGradients(int n,double a,double b)
{
    auto start1 = high_resolution_clock::now();

    FILE *ft=fopen("autodiffgradients.txt","w");
    for(double x= a;x<=b;x+=(b-a)/n)
    {
        dual xx=x;
        double dudx = derivative(af, wrt(xx), at(xx));   // evaluate the derivative du/dx
        fprintf(ft,"%20.10lf %20.10lf\n",x,dudx);
    }
    fclose(ft);
    auto stop1 = high_resolution_clock::now();
    auto duration1 = duration_cast<microseconds>(stop1 - start1);
    printf("Time Passed: %ld \n",duration1);
}

int main()
{
    const double a=-10;
    const double b= 10;
    const int n = 100000;
    printRealGradients(n,a,b);


    printFiniteGradients(n,a,b);
    printAutodiffGradients(n,a,b);
    return 0;
}
